from collections import Counter
from copy import copy
import re

from .base import Name, Taxon
from taxonome.tracker import noop_tracker, prepare_tracker
from taxonome.taxa.name_selector import TerminalNameSelector, TryAnotherName
from taxonome.utils.qgram import QGramIndex
select_name = TerminalNameSelector()

def _filter_non_auth(possnames, name):
    """Filter out name pairs with 'non X' authority conflicts. Used in .select()
    below."""
    return [(n,an) for n,an in possnames if not n.authority.non_match(name.authority)]

class TaxaResource:
    """A base class for classes holding indexed collections of taxa.
    
    Not designed to be instantiated."""
    _nameselector = select_name
    
    def resolve_name(self, name):
        """Needs to be overridden by subclass. Should return a list of 2-tuples,
        (matching_name, accepted_name), matching the name given. The name given
        may be a string or a Name object."""
        raise NotImplementedError
    
    def get_by_accepted_name(self, name):
        """Needs to be overridden by subclass. Called with a Name object known
        to be an accepted name in this resource. Should return the corresponding
        taxon."""
        raise NotImplementedError
    
    def wildcard_resolve_name(self, name_pattern):
        """May be defined by subclasses to allow wildcard matching of names using
        the common * wildcard, and any extra pattern features webservices etc.
        may support. Returns 2-tuples like resolve_name. name_pattern should be
        a string."""
        raise NotImplementedError
    
    def resolve(self, key, wildcard=False):
        """Return a set of taxa matching a name. The name can be an unqualified
        name as a string, a qualified (with authority) Name object, or a Taxon
        object (which is treated as a shorthand for passing its .name property).
        
        If wildcard is True, attempt to match names to a string using the *
        wildcard. This may raise NotImplementedError if the data source
        doesn't support it.
        """
        nameresolver = self.wildcard_resolve_name if wildcard \
                            else self.resolve_name
        return {self.get_by_accepted_name(an) for n, an in nameresolver(key)}
    
    def select(self, name, strict_authority=False, upgrade_subsp="nominal",
                    fuzzy=True, prefer_accepted='noauth', nameselector=None,
                                                        tracker=noop_tracker):
        """Select a taxon by name. If there is more than one match, an
        interactive prompt will ask the user to choose.
        
        Parameters:
            name : Name or str
              The name (qualified or unqualified) to look for
            strict_authority : bool
              If True, names will be disregarded if their authority does not match
              that given (with a qualified name). If False, the authority will be
              used to distinguish homonyms, but will be overlooked if none of the
              names found have matching authority.
            upgrade_subsp : str
              If 'none', subspecies and varieties will only be matched to identical
              names. Otherwise, subspecies and varieties may be matched to their
              parent species if no exact match is found. 'nominal' will cause only
              nominal subspecies/varieties to be upgraded (e.g. *Vicia faba* var. *faba*
              --> *Vicia faba*), while 'all' allows any subspecies/variety to upgrade.
            fuzzy : bool
              If True, fuzzy matching will be used with the specified name where
              possible.
            prefer_accepted : str
              'all' will resolve ambiguities by choosing the accepted name where
              possible. 'noauth' does this only where the given name doesn't have an
              authority. 'none' will always ask the user.
            nameselector : NameSelector instance
              Defines how to choose between alternative matches. See
              :class:`taxonome.taxa.name_selector.NameSelector`. The default
              name selector is the _nameselector attribute on this collection.
            tracker :
              A tracker object - see details in :mod:`taxonome.tracker`.
        
        Returns:
            A taxon object if a match is found in this dataset. Raises KeyError
            if no match is found.
        """
        nameselector = nameselector or self._nameselector
        possnames = self.resolve_name(name)
        
        # Overlook authority?
        if not possnames and not strict_authority and hasattr(name, "plain"):
            tracker.name_event(name, "authority overlooked")
            possnames = _filter_non_auth(self.resolve_name(name.plain), name)
        
        # Upgrade minor taxa?
        if not possnames and isinstance(name, Name) and \
                name.subnames and (upgrade_subsp=="all"\
                or (upgrade_subsp=="nominal" and name.is_nominal_subsp)):
            pn = name.parent
            tracker.name_transform(name, pn, "upgraded subspecies")
            possnames = self.resolve_name(pn)
            if (not possnames) and (not strict_authority):
                possnames = _filter_non_auth(self.resolve_name(pn.plain), pn)
        
        # Fuzzy name matching?
        if not possnames and fuzzy:
            try:
                possnames = self.fuzzy_resolve_name(name)
                tracker.name_event(name, "tried fuzzy match")
            except (AttributeError, NotImplementedError):
                pass
            else:
                if not possnames and not strict_authority and hasattr(name, "plain"):
                    possnames = _filter_non_auth(self.fuzzy_resolve_name(name.plain), name)
        
        # Pick accepted name by default?
        if prefer_accepted=='all' or (prefer_accepted=='noauth' and \
                                (isinstance(name, str) or not name.authority)):
            acc_count = 0
            for n, an in possnames:
                if n==an:
                    acc_count += 1
                    an1 = an
            if acc_count == 1:
                tracker.name_transform(name, an1, "preferring accepted name")
                return self.get_by_accepted_name(an1)
        
        try:
            n, an = nameselector(possnames, name, tracker=tracker)
            return self.get_by_accepted_name(an)
        except TryAnotherName as e:
            return self.select(e.newname, strict_authority=strict_authority,
                                upgrade_subsp=upgrade_subsp, fuzzy=fuzzy,
                                tracker=tracker)

class TaxonSet(TaxaResource):
    """This allows storing a collection of taxa, along with synonymy
    and easy lookup by name (whether or not that is qualified with
    an authority.
    
    Looking up synonyms that have been added returns the taxon, but
    looping through it will ignore synonyms.
    
    Usage example::
    
      mygenus = TaxonSet()
      mygenus.add(Taxon("Lablab purpureus","(L.) Sweet"))
      mygenus.add_synonym(Name("Dolichos lablab", "L."),\\
                        Name("Lablab purpureus","(L.) Sweet") )
      mygenus.add_synonym(Name("Vigna aristata", "Piper"),\\
                        Name("Lablab purpureus","(L.) Sweet") )
                        
      # Unqualified name:
      "Lablab purpureus" in mygenus       # -> True
      "Dolichos lablab" in mygenus        # -> True
      "Lablab vulgaris" in mygenus        # -> False
      "Dolichos lablab L." in mygenus     # -> False (Don't add authority)
      mygenus.resolve("Lablab purpureus") # -> List of possible Taxon() objects
      mygenus.resolve("Dolichos lablab")  #    (will usually only have one item)
      mygenus["Lablab purpureus"]         # -> A Taxon() object, so long as there
                                          #    is exactly one match.
    
      # Qualified name:
      Name("Lablab purpureus","(L.) Sweet") in mygenus # -> True
      Name("Lablab purpureus","Sweet") in mygenus      # -> True
      mygenus[Name("Dolichos lablab","L.")]  # Returns a Taxon() object
      mygenus[Taxon("Dolichos lablab","L.")] # Same - Only uses the taxon's
                                             # accepted name
    """
    type_err_msg = "Don't know how to look for type {0}"
    def __init__(self):
        self.names = {}
        self.taxa = {}
        self.qgram_index = QGramIndex()
        
    def add(self, taxon):
        """Add a taxon. This will be indexed by its name, and any synonyms
        attached to the taxon object.
        
        Adding a taxon with the same name as an existing taxon will replace it.
        An accepted name will replace the same qualified name as a synonym for
        another taxon, although the synonym will still be in the .othernames
        property of the relevant taxon.
        """
        if not taxon.name in self:
            self._add_qname(taxon.name)
        elif self.select(taxon.name).name != taxon.name:
            self._del_qname(taxon.name)
            self._add_qname(taxon.name)
        #if taxon.name in self:
            #if self[taxon.name].name == taxon.name:
                #raise KeyError("Taxon already added")
            #else:
                #self._del_qname(taxon.name)
        self.taxa[repr(taxon.name)] = taxon
        #self._add_qname(taxon.name)
        for synonym in taxon.othernames:
            if not synonym in self:
                self._add_qname(synonym, taxon.name)
        
    def _add_qname(self, qname, acceptedname=None):
        """Add a qualified name to the index. To add a synonym, specify an
        acceptedname to which it maps. The acceptedname should refer to a taxon
        which is in the set.
        """
        if not acceptedname:
            acceptedname = qname
        if not qname.plain in self.names:
            self.names[qname.plain] = []
            self.qgram_index.add(qname.plain)
        self.names[qname.plain].append((qname, acceptedname))
        
    def _del_qname(self, qname):
        """Remove a qualified name from the index. Taxa with no remaining
        names in the index will still take up space, but cannot easily be
        accessed, so you should ensure that at least one name remains for each.
        
        Returns the accepted name to which it mapped.
        """
        if not qname.plain in self.names:
            raise KeyError
        for namematch in self.names[qname.plain]:
            if namematch[0] == qname:
                self.names[qname.plain].remove(namematch)
                if not self.names[qname.plain]:
                    del self.names[qname.plain]
                    self.qgram_index.discard(qname.plain)
                return namematch[1]
        raise KeyError(qname)
    
    def __contains__(self, taxon):
        if isinstance(taxon, str):
            return taxon in self.names
        elif isinstance(taxon, Name):   # Qualified name
            if taxon.plain in self.names:
                return any([taxon == n for n,t in self.names[taxon.plain] ])
        elif isinstance(taxon, Taxon):
            return self.__contains__(taxon.name) # Redo using name
        else:
            raise TypeError(self.type_err_msg.format(type(taxon)))
    
    def resolve_name(self, key):
        """Return a list of 2-tuples, (name, accepted_name), matching the given
        name. The name can be an unqualified name as a string or a qualified
        (with authority) Name object.
        """
        if isinstance(key, Name):  # Qualified name
            return [(n, an) for n, an in self.names.get(key.plain, []) if key == n]
        # Unqualified name as string
        return self.names.get(key, [])
    
    def fuzzy_resolve_name(self, key):
        """Return a list of 2-tuples as for :meth:`resolve_name`, but including
        fuzzy matches to the specified name.
        """
        rawkey = key.plain if isinstance(key, Name) else key
        allnames = []
        for similarname, score in self.qgram_index.matches(rawkey):
            allnames.extend(self.names[similarname])
        if isinstance(key, Name):
            return [(n, an) for n, an in allnames if n.authority==key.authority]
        return allnames
    
    def wildcard_resolve_name(self, name_pattern):
        """Return a list of 2-tuples as for :meth:`resolve_name`, but matching
        a wildcard pattern in the Unix shell style. See the docs for the fnmatch
        module for more details."""
        import fnmatch
        pattern_re = re.compile(fnmatch.translate(name_pattern))
        allnames = []
        for plainname in self.names:
            if pattern_re.match(plainname):
                allnames.extend(self.names[plainname])
        return allnames
    
    def get_by_accepted_name(self, key):
        """Return the taxon identified by an accepted name."""
        return self.taxa[repr(key)]
    
    def __getitem__(self, key):
        """The simplest retrieval option. Will return a single match, or raise
        an error if there is more than one match."""
        matches = self.resolve(key)
        if len(matches) == 1:
            return matches.pop()
        elif len(matches) < 1:
            raise KeyError("No matching taxa found for '%s'" % key)
        else:
            raise KeyError("Multiple matches found for '%s'" % key)
    
    def __setitem__(self, key, value):
        if key in self:
            key = self[key].name
            self.taxa[repr(key)] = value
        else:
            self.add(value)
            if key != value.name:
                self._add_qname(key, value.name)

    def __delitem__(self, key):
        if isinstance(key, str):
            if len(self.names[key]) > 1:
                raise KeyError("More than one match for name")
            name = self.names[key][0]
            del self.taxa[repr(name[1])]
            del self.names[key]
        elif isinstance(key, Name):
            # Qualified name
            accname = self._del_qname(key)
            if accname != key:
                self._del_qname(accname)
            for synonym in self.taxa[repr(accname)].othernames:
                try:
                    synonym_maps_to = self._del_qname(synonym)
                    if synonym_maps_to != accname:
                        # Oops, we removed a name that had been overwritten. Put it back.
                        self._add_qname(synonym, synonym_maps_to)
                except KeyError:
                    pass
            del self.taxa[repr(accname)]

        elif isinstance(key, Taxon):
            self.__delitem__(key.name)
        else:
            raise TypeError(self.type_err_msg.format(type(taxon)) )
        
    def __iter__(self):
        return iter(self.taxa.values())
            
    def __len__(self):
        return len(self.taxa)
    
    def __repr__(self):
        return "<{0}: {1} taxa with {2} names>".format(self.__class__.__name__,\
                        len(self.taxa), len(self.names))

def combine_dicts(d1, d2):
    """Default function for combining info dicts. It gives the union of two
    dicts, while preserving multiple values with matching keys. If the
    corresponding values are both dicts, they are likewise combined; otherwise
    values are placed in a list."""
    res = copy(d1)
    for k, v in d2.items():
        if k not in res:
            res[k] = v
        elif isinstance(res[k], dict) and isinstance(v, dict):
            res[k] = combine_dicts(res[k], v)
        elif isinstance(res[k], set) and isinstance(v, set):
            res[k].update(v)
        elif isinstance(res[k], list):
            if isinstance(v, list):
                if v != res[k]:
                    res[k].extend(v)
            else:
                res[k].append(v)
        else:
            res[k] = [res[k], v]
    return res

def build_matched_taxonset(streaming_match_taxa, merge_info=combine_dicts):
    results = TaxonSet()
    
    for taxon, t2 in streaming_match_taxa:
        try:
            restaxon = results.select(t2.name)
        except KeyError:
            taxon = copy(taxon)
            taxon.name = t2.name
            taxon.othernames = set()
            results.add(taxon)
        else:
            restaxon.info = merge_info(restaxon.info, taxon.info)
            restaxon.distribution.update(taxon.distribution)
            restaxon.individuals.extend(taxon.individuals)
    
    return results

def match_taxa(taxa, target, merge_info=combine_dicts,
               tracker=noop_tracker, **kwargs):
    """Match one set of taxa against another by name.
    
    Parameters:
        taxa : iterable
          The taxa to be matched
        target : TaxonSet
          The taxa whose names we want to match to.
        merge_info : function
          A function which takes two dictionaries as arguments, and returns a
          dictionary. Called with the .info parameters of multiple taxa which are
          mapped to the same name.
        tracker :
          A tracker object (see :mod:`taxonome.tracker`) or tuple of trackers to
          record the name matching process.
        strict_authority, upgrade_subsp, fuzzy, prefer_accepted, nameselector :
          Control the matching procedure; see :meth:`TaxonSet.select` for details.
    
    Returns:
        A TaxonSet, containing copies of the input taxa, with the .name
        attribute replaced with the name of the corresponding taxon from target_ts,
        .othernames cleared, and .info combined with taxa mapping to the same name.
        
        Taxa which fail to match send a 'no match' event to the tracker.
    """
    matching = streaming_match_taxa(taxa, target, tracker=tracker, **kwargs)
    return build_matched_taxonset(matching, merge_info=merge_info)

def streaming_match_taxa(taxa, target, tracker=None, cache_hwm=1000,
                                                    cache_lwm=800, **kwargs):
    """Similar to match_taxa, but suitable for handling large amounts of data.
    
    This is a generator, yielding pairs of (input_taxon, matched_taxon). The same
    matched taxon may be yielded more than once.
    
    This keeps a cache of commonly seen names; when the number of names in the
    cache reaches cache_hwm (default 1000), it is cut to cache_lwn (default 800).
    These default sizes are quite conservative - they could probably go at least
    an order of magnitude higher.
    """
    tracker = prepare_tracker(tracker)
    cache = {}
    names_seen = Counter()
    
    for taxon in taxa:
        tracker.reset()
        tracker.start_taxon(taxon)
        names_seen[taxon.name] += 1
        
        # Look up name in cache
        try:
            t2 = cache[taxon.name]
        except KeyError:
            pass
        else:
            if t2:
                tracker.name_transform(taxon.name, t2.name, "reapplied previous match")
                yield taxon, t2
            else:
                tracker.name_event(taxon.name, "no match")
            continue
        
        # Not in cache - look in the target dataset
        try:
            t2 = target.get_by_accepted_name(taxon.name)
        except KeyError:
            try:
                t2 = target.select(taxon.name, tracker=tracker, **kwargs)
            except KeyError:
                t2 = None
        else:
            tracker.name_transform(taxon.name, t2.name, "exact match")
        
        # Cache matched taxon, or None if no match
        cache[taxon.name] = t2
        
        if t2 is None:
            tracker.name_event(taxon.name, "no match")
            continue
        
        yield taxon, t2
        
        # Limit cache size
        if len(cache) > cache_hwm:
            names_seen = Counter(dict(names_seen.most_common(cache_lwm)))
            cache = {name:cache[name] for name in names_seen}
        
    tracker.reset()

def run_match_taxa(taxa, target, **kwargs):
    """Run the matching process without collecting the results - suitable
    for use with big datasets.
    """
    for taxonpair in streaming_match_taxa(taxa, target, **kwargs):
        pass



#!/usr/bin/python3
# -*- coding: utf-8 -*-
import sys
sys.path[0] = "." # Ugly hack so we can import ourself.
from taxonome import col

print("Commands: resolve, equivalent, map, quit, help")
while True:
    cmd, dummy, info = input(":").partition(" ")
    cmd = cmd.lower()
    if cmd.startswith("quit") or cmd.startswith("exit"):
        break
    elif cmd.startswith("help"):
        print("Examples:")
        print('  resolve "Dolichos lablab"')
        #print('  equivalent "Dolichos lablab" "Lablab purpureus"')
        #print("  map species_listA.csv species_listB.csv")
        print("  quit")
    if cmd.startswith("resolve"):
        print(col.resolve(info.strip(" \"")))



#!/usr/bin/python3
# -*- coding: utf-8 -*-
import re

def grabchunks(text, opener, closer):
    """
    Get 'chunks' as marked by the specified opener and closer. Only the first
    level of chunks will be returned. These may contain similar chunks, which
    can be extracted by running the function again.
    
    This is a generator: to get an indexable list of the chunks, use
    list(grabchunks(...))
    
    Example usage:
    foo = "green <eggs <and>> multicoloured <spam>."
    for chunk in grabchunks(foo, "<",">"):
        print(chunk)
    # eggs <and>
    # spam
    """
    assert opener != closer, "Opener and closer must be different"
    if opener not in text:
        return
    openpoints = [m.start() for m in re.finditer(re.escape(opener), text)]
    closepoints = [m.start() for m in re.finditer(re.escape(closer), text)]
    while closepoints[0] < openpoints[0]:
        closepoints.pop(0) # Ignore any closers before first opener
    chunkstart = openpoints.pop(0) + len(opener) # Exclude opening tag
    inlevel = 1
    while (closepoints or openpoints):
        if openpoints and (openpoints[0] < closepoints[0]):
            if inlevel <= 0:
                chunkstart = openpoints.pop(0) + len(opener)
                inlevel = 0 # Ensure it doesn't become negative
            else:
                openpoints.pop(0)
            inlevel += 1
        else:
            chunkend = closepoints.pop(0)
            inlevel -= 1
            if inlevel == 0:
                yield text[chunkstart:chunkend]
      
def splitwithout(text, spliton, containermarks = ['"']):
    """Splits text in a similar manner to text.split(spliton), but ignoring
    spliton where it occurs between start and end markers. E.g. 
    'a,<b,c>' could be split into ['a', '<b,c>'], rather than ['a', '<b', 'c>'].

    You can specify any number of containers, and the text will only be split 
    where the delimiter occurs outside all of them.
    For each container, specifying a single string will apply it as a toggle,
    as in "simple quotation marks", while a 2-tuple can contain separate start
    and end markers (e.g. brackets), which can be nested: (1 (2(...))).
    
    Example usage:
    foo = 'Green, "Eggs, Spam", etc.'
    splitwithout(foo, ',')
    # ['Green', ' "Eggs, Spam"', ' etc.']

    bar = "Do | {{Re | {{Mi | Fa }}}} | Sol | <La | Ti> | Do"
    splitwithout(bar,'|', [('{{','}}'), ('<','>')])
    # ['Do ', ' {{Re | {{Mi | Fa }}}} ', ' Sol ', ' <La | Ti> ', ' Do']"""
    posssplitpoints = [m.start() for m in re.finditer(re.escape(spliton), text)]
    for containers in containermarks:
        if isinstance(containers, str): # Same opener and closer
            opener = containers
            closer = None
        else:   # Different opener and closer.
            opener, closer = containers
        openpoints = [m.start() for m in re.finditer(re.escape(opener), text)]
        splitpoints = []
        if closer and (closer != opener): #Separate opener and closer (can be nested)
            closepoints = [m.start() for m in re.finditer(re.escape(closer), text)]
            inlevel = 0
            prevpoint = 0
            for point in posssplitpoints:
                while openpoints and (openpoints[0] < point):
                    openpoints.pop(0)
                    inlevel += 1
                while closepoints and (closepoints[0] < point):
                    closepoints.pop(0)
                    inlevel -= 1
                inlevel = max(inlevel, 0) #Stop it becoming negative.
                if inlevel == 0:
                    splitpoints.append(point)
        else: # Same opener and closer
            within = 0
            prevpoint = 0
            for point in posssplitpoints:
                togglecount = sum(x < point for x in openpoints)
                openpoints = openpoints[togglecount:]
                within = (togglecount + within) % 2
                if not within:
                    splitpoints.append(point)
        posssplitpoints = splitpoints
        
    sections = []
    prevpoint = 0
    for point in splitpoints:
        sections.append(text[prevpoint:point])
        prevpoint = point + len(spliton)
    return sections + [text[prevpoint:]]
    
            



#!/usr/bin/python3
# -*- coding: utf-8 -*-
from urllib.request import Request, urlopen
from urllib.parse import urlencode
import re
import os.path

from taxonome import config, Taxon, Name, TaxonSet
from taxonome.utils import structxt, fshelf
from taxonome.taxa.base import TaxonShelf

class NoInfoError(Exception):
    pass

USER_AGENT = config['main']['user-agent']

CACHE_FILE = os.path.join(config['cache']['location'], 'wikipedia_cache')
EXPIRY = datetime.timedelta(days=config.getint('cache', 'expiry'))

#cache = fshelf.ForgetfulShelf(CACHE_FILE, expiry=EXPIRY)
titlecache = fshelf.ForgetfulShelf(CACHE_FILE+"_titles", expiry=EXPIRY)
cache = TaxonShelf(CACHE_FILE)

RAW_BASE = "http://en.wikipedia.org/w/index.php?"
PAGE_BASE = "http://en.wikipedia.org/wiki/"
Redirect_RE = re.compile(r'#REDIRECT ?\[\[(.+?)\]\]',re.I) # I=Ignore Case
Smalltext_RE = re.compile(r'<small>(.+?)</small>', re.I)
RANKS = ['trinomial','binomial','genus','familia','ordo','classis','phylum','regnum']

def get_page(title, follow_redirects=True):
    """Gets the raw wikicode of a page, given the title."""
    query = urlencode({'title':title.replace(" ","_"), 'action':'raw'})
    page = urlopen(Request(RAW_BASE + query, headers={'ua':USER_AGENT}))
    content = page.read().decode('utf-8')
    if follow_redirects and content.upper().startswith("#REDIRECT"):
        new_title = Redirect_RE.match(content).group(1)
        titles, page = get_page(new_title)
        return titles + [title], page
    else:
        return [title], content
    
def _get_taxobox(page):
    """Cuts the taxobox out of a page, and translates its parameters
    into a dictionary."""
    for template in structxt.grabchunks(page, '{{','}}'):
        if template.strip().lower().startswith('taxobox'):
            break
    else:
        raise NoInfoError("No taxobox found")
    slots = structxt.splitwithout(template, '|', [('{{','}}'),('[[',']]')])
    params = {}
    for slot in slots[1:]:
        name, dummy, value = slot.partition("=")
        params[name.strip().lower()] = value.strip()
    return params

def get_taxobox(name):
    """Look up a name, and return the parameters in its taxobox as a
    dictionary."""
    title, page = get_page(name)
    return title, _get_taxobox(page)
    
def _munge_synonyms(wikicode):
    synlines = wikicode.split("*")
    if len(synlines) <= 1:
        wikicode = wikicode.replace("<br>","<br/>")
        synlines = wikicode.split("<br/>")
    synonyms = []
    for synline in synlines:
        synline = synline.strip()
        if not synline:
            continue
        if "<small>" in synline: # small tags are often used for authors
            name = _strip_wikicode(synline[:synline.find("<small>")] )
            auth = _strip_wikicode(Smalltext_RE.search(synline).group(1) )
            synonyms.append(Name(name, auth))
            continue
        italicbits = synline.split("''")
        if italicbits[0].strip() != "":
            continue # Name should start italic
        name = " ".join(italicbits[1:-1]) # This allows for var./subsp.
        auth = _strip_wikicode(italicbits[-1])
        synonyms.append(Name(name, auth))
    return synonyms

    
def _strip_wikicode(text):
    # Tidy up links, including piped links
    text = re.sub(r"\[\[(.+?\|)?(.+?)\]\]",r"\2",text)
    while "<ref" in text: # Cut references out.
        refstart = text.find("<ref")
        if "</ref>" in text:
            refend = text.find("</ref>") + len("</ref>")
        else:
            refend = text.find("/>",refstart) + 2
        text = text[:refstart] + text[refend:]
    return text.replace("'''","").replace("''","").strip()

def resolve(qname):
    if qname in cache:
        return cache[qname]
    if isinstance(qname, str) and qname in titlecache:
        return [cache[titlecache[qname]]]
    if hasattr(qname, "name"):
        name = qname.name
    else:
        name = qname
    title, taxobox = get_taxobox(name)
    for rank in RANKS:
        if rank in taxobox:
            accname = _strip_wikicode(taxobox[rank])
            authkey = rank + "_authority"
            if authkey in taxobox:
                auth = _strip_wikicode(taxobox[authkey])
            else:
                auth = ""
            break
    else:   # No taxon name found.
        return None
    wp_taxon = Taxon(accname, auth)
    wp_taxon.url = PAGE_BASE + title[0].replace(" ","_")
    if 'synonyms' in taxobox:
        wp_taxon.othernames = _munge_synonyms(taxobox['synonyms'])
    cache.add(wp_taxon)
    for pagetitle in title:
        if not wp_taxon.hasname(pagetitle):
            titlecache[pagetitle] = wp_taxon.name
    if hasattr(qname, "name"):
        if wp_taxon.hasname(qname):
            return wp_taxon
        return None      # Name found, but qualifier didn't match, so don't return.
    return [wp_taxon]



# -*- coding: utf-8 -*-
"""Handle geographic regions, which can be nested within one another.

This automatically loads the TDWG regions, together with indexes of
names and ISO country codes (see tdwg_load.py for details of the data
sources). They are accessible via the module variables ``tdwg``, ``names`` and ``iso``,
as well as specific TDWG levels in ``tdwg_levels[1]`` - ``tdwg_levels[4]``.
The variable ``index`` combines these three sets of names for convenience.

Example usage::

  regions.tdwg["IND"]     # TDWG level 3 region (most of India)
  regions.names["India"]  # The country
  i = regions.ISO["IN"]   # The country (ISO code)
  regions.index["India"]  # .index includes all codes and names
  print(i.children)       # Show immediate subregions

  for subregion in i:     # Gets all subregions, down to the smallest
      print subregion.name

  ## Test if a region is in a larger region:
  "NET" in regions.names["Europe"]  # True: The Netherlands is in Europe
  "NET" in regions.tdwg["14"]       # False (14 is Eastern Europe, TDWG level 2)

  for region in regions.tdwg_levels[1]:
      print(region.name)     # tdwg_levels has a set of the regions at each level

You can also use the add to the hierarchy, or create your own (e.g. for a much
smaller area). See the :class:`Map` class.
"""
import logging
try:
    from unidecode import unidecode
except ImportError:
    def unidecode(input): return input

from taxonome.config import config
from taxonome.utils.digraph import DiGraph
                    
class Map(DiGraph):
    subregions = DiGraph.successors
    is_subregion = DiGraph.is_successor
    
    superregions = DiGraph.predecessors
    immediate_superregions = DiGraph.immediate_predecessors
    
    def add_region(self, region, parent=None):
        self.nodes.add(region)
        if parent:
            self.edges.add(parent, region)
            
    def includes(self, region, targets):
        """Test whether a region in the map includes any of a collection of
        regions, such as an organism's distribution. E.g.
        
        africa = tdwg["2"]
        world.includes(africa, elephant.distribution)
        
        The return value is the set of matching subregions: this can be treated
        as a boolean, since it is empty if there is no overlap.
        """
        subregions = set(self.subregions(region)) | {region}
        return subregions.intersection(targets)
        
    def immediate_subregions(self, region):
        """Return only the immediate child regions of the given region."""
        return set(self.outgoing(region))


class CombinedIndex(list):
    """A sequence of dictionaries: will retrieve an item
    from the first one it is found in.
    """
    def __contains__(self, key):
        return any(key in index for index in self)
    
    def __getitem__(self, key):
        for index in self:
            if key in index:
                return index[key]
        raise KeyError(key)

def _find_tdwg_level(region):
    for n, contents in enumerate(tdwg_levels):
        if region in contents:
            return n

def find_tdwg(name, level=3, index=None):
    """Find the set of TDWG regions at a given level corresponding to a given
    name.
    
    Parameters:
      name : str
        The name or code to search for.
      level : int
        The target TDWG level (1-4). Defaults to 3.
      index :
        The index to use for looking up names/codes. Defaults to names. Others
        available in this module are: tdwg, ISO (two-letter ISO country codes, e.g
        NL) and combinedindex (which checks tdwg, names and ISO).
    
    Returns:
      A set of TDWG codes; either all the regions contained within the specified
      region, or the single region which contains or matches the specified region.
    
    Raises: KeyError if the name/code is not found.
    """
    if not world: load_tdwg()
    
    if index is None:
        index = names
    
    r = index[name]
    rlevel = _find_tdwg_level(r)
    if rlevel == level:
        return {r[1]}
    
    elif rlevel is None:   # Extra regions without a TDWG code.
        out = set()
        for n,t in world.immediate_subregions(r):
            out.update(find_tdwg(t, level=level, index=tdwg))
        if not out:
            for n,t in world.immediate_superregions(r):
                out.update(find_tdwg(t, level=level, index=tdwg))
        return out
    
    elif rlevel < level:  # Larger regions
        return {sr[1] for sr in world.subregions(r) if sr in tdwg_levels[level]}
    
    # Smaller regions
    return {gr[1] for gr in world.superregions(r) if gr in tdwg_levels[level]}
    
def tdwgise(namegroup, level=3, index=None):
    """Transform an iterable of names to a set of TDWG regions at a particular
    level.
    
    Note that, if the input specifies a larger region than the required tdwg
    level, the output will include all subregions of that region. This may lead
    to some inaccuracies - e.g. a species recorded as occurring in 'Africa'
    will appear to occur all over Africa.
    
    Parameters: The same as :func:`find_tdwg`, except that the first parameter
    is an iterable of names.
    
    Returns:
      A pair of sets: the matching TDWG codes at the specified level, and any
      names which could not be matched.
    """
    out, notfound = set(), set()
    for name in namegroup:
        try:
            t = find_tdwg(name, level, index)
        except KeyError:
            notfound.add(name)
        else:
            out.update(t)
    return out, notfound
    
def load_tdwg():
    """Call this function to load the TDWG set of world regions.
    
    If the 'load-tdwg-regions' config value is True, this will be called when
    the module is imported.
    """
    global world, names, tdwg, ISO, combinedindex, tdwg_levels
    from . import tdwg_data
    world = Map()
    names = {}
    tdwg = {}
    ISO = {}
    combinedindex = CombinedIndex([tdwg, names, ISO])
    
    for code, name, isocode, parent, notes in tdwg_data.data:
        newregion = (name, code)
        if parent == "!!WORLD":
            parent = None
        else:
            parent = tdwg[parent]
        world.add_region(newregion, parent)
        
        # Add to indices
        tdwg[code] = newregion
        if name not in names:
            names[name] = newregion
            # Strip off accents for more reliable matching
            names[unidecode(name)] = newregion
        if isocode and isocode not in ISO:
            ISO[isocode] = newregion
        
    for name, parent, children in tdwg_data.extra_countries:
        newregion = (name, None)
        if parent:
            parent = tdwg[parent]
        world.add_region(newregion, parent)
            
        for childcode in children:
            child = tdwg[childcode]
            world.edges.add(newregion, child)
            if parent:   # Cut any direct links from the old parent to children
                world.edges.discard(parent, child)
        names[name] = newregion
        
    for name, origname in tdwg_data.extra_names:
        names[name] = names[origname]
    for name, tcode in tdwg_data.extra_names_tdwg:
        names[name] = tdwg[tcode]
    for code, name in tdwg_data.extra_ISO.items():
        ISO[code] = names[name]
        
    tdwg_levels = [set()]
    for levelcodes in tdwg_data.tdwg_by_level[1:]:
        tdwg_levels.append({tdwg[code] for code in levelcodes})
        
world = None   # Test for this to see if TDWG data is loaded.
if config.getboolean('main', 'load-tdwg-regions'):
    load_tdwg()



import io

from .test_taxa import LiquoriceTestBase
from taxonome.taxa import file_jsonlines

class JsonlinesTest(LiquoriceTestBase):
    def test_roundtrip(self):
        f = io.StringIO()
        file_jsonlines.save_taxa(f, self.ts)
        f.seek(0)
        newts = file_jsonlines.load_taxa(f)
        
        self.assertEqual(len(newts), 2)
        self.assertIn("Glycyrrhiza glabra", newts)
        self.assertIn(self.liquorice.name, newts)
        homonyms = newts.resolve_name("Glycyrrhiza glandulifera")
        self.assertEqual(len(homonyms), 2)



# -*- coding: utf8 -*-
from taxonome.taxa.base import Name
import csv
from collections import defaultdict

class NoopTracker:
    """This tracker doesn't do anything."""
    name_event = name_transform = start_taxon = reset = lambda *args: None

noop_tracker = NoopTracker()

class MultiTracker:
    """Holds a list of trackers, and calls the relevant methods of each of them
    when its methods are called."""
    def __init__(self, *args):
        self._trackers = args
    
    def __getattr__(self, name):
        methods = [getattr(t, name) for t in self._trackers if hasattr(t, name)]
        def dispatch(*args, **kwargs):
            for m in methods:
                m(*args, **kwargs)
        return dispatch

def prepare_tracker(tracker):
    if isinstance(tracker, (list, tuple)):
        return MultiTracker(*tracker)
    if tracker is None:
        return noop_tracker
    return tracker

def coroutine(func):
    """Decorator to prime coroutines"""
    def start(*args,**kwargs):
        cr = func(*args,**kwargs)
        next(cr)
        return cr
    return start

RESET = object()
@coroutine
def csvchainwriter(writer):
    """Coroutine to write lines to a csv file, incrementing a chain number until
    it's sent a reset signal."""
    chain = 0
    while True:
        line = (yield)
        if line is RESET:
            chain = 0
        else:
            writer.writerow((chain,) + tuple(line))
            chain += 1

def _flatten_name(name):
    if isinstance(name, Name):
        return name.plain, str(name.authority)
    return name, None

class CSVTracker(NoopTracker):
    """This writes events to a CSV file."""
    def __init__(self, fileobj, header=True):
        writer = csv.writer(fileobj)
        if header:
            writer.writerow(["sequence", "From name", "From authority", "To name", "To authority", "Event"])
        self.push = csvchainwriter(writer).send
        
    def name_event(self, name, event):
        self.push(_flatten_name(name) + (None, None, event))
    
    def name_transform(self, name, newname, event):
        self.push(_flatten_name(name) + _flatten_name(newname) + (event,))
    
    def reset(self):
        self.push(RESET)

class CSVListMatches(NoopTracker):
    """Writes the original name and final match to a CSV file."""
    fromname = None
    toname = None
    
    def __init__(self, fileobj, header=True):
        self.writer = csv.writer(fileobj)
        if header:
            self.writer.writerow(["Name","Authority","Matched to",None])
        
    def name_event(self, name, event):
        if not self.fromname:
            self.fromname = name
        self.toname = None
    
    def name_transform(self, name, newname, event):
        if not self.fromname:
            self.fromname = name
        self.toname = newname
    
    def reset(self):
        if self.fromname:
            self.writer.writerow(_flatten_name(self.fromname) + _flatten_name(self.toname))
        self.fromname = None
        self.toname = None

class Counter(NoopTracker):
    """Count the number of names seen so far. Callback with the number after
    every n (default 10).
    """
    def __init__(self, callback, every=10):
        self.callback = callback
        self.every = every
        self.n = 0
    
    def start_taxon(self, tax):
        self.n += 1
        if self.n % self.every == 0:
            self.callback(self.n)

class EventCounter:
    def __init__(self):
        self.started = 0
        self.events = defaultdict(int)
    
    def start_taxon(self, tax):
        self.started += 1
    
    def name_event(self, name, event):
        self.events[event] += 1
    
    def name_transform(self, name, newname, event):
        self.events[event] += 1

class CSVTaxaTracker(NoopTracker):
    """Produces a list of the existing taxa data with the new names."""
    taxon = None
    newname = None
    
    def __init__(self, fileobj, fieldnames, header=True, include_unmatched=False):
        fieldnames = ["Name", "Authority", "Original name", "Original authority"]+fieldnames
        self.writer = csv.DictWriter(fileobj, fieldnames, extrasaction='ignore')
        if header:
            self.writer.writeheader()
        
        self.include_unmatched = include_unmatched
    
    def start_taxon(self, tax):
        self.taxon = tax
    
    def name_transform(self, name, newname, event):
        self.newname = newname
    
    def reset(self):
        if self.taxon and (self.newname or self.include_unmatched):
            d = dict(self.taxon.info)
            d['Original name'], d['Original authority'] = _flatten_name(self.taxon.name)
            d['Name'], d['Authority']= _flatten_name(self.newname)
            self.writer.writerow(d)
        
        self.taxon = None
        self.newname = None



import unittest

from taxonome.taxa.author import Author, Authority

class AuthorityTests(unittest.TestCase):
    def test_simple_authority(self):
        a = Authority("L.")
        self.assertEqual(a.main[0], Author("L."))
        assert a.ex == a.basauth == a.non == ()

    def test_complex_authority(self):
        a = Authority("(Smith) Gray & Presl. ex. L. Stev.")
        self.assertEqual(a.basauth[0], Author("Smith"))
        self.assertEqual(a.main, (Author("Gray"), Author("Presl.")))
        self.assertEqual(a.ex[0], Author("L. Stev."))
        self.assertEqual(str(a), "(Smith) Gray & Presl. ex L. Stev.")
        
        self.assertGreater(a.fuzzy_cmp(Authority("Gray and Presley")), 0.5)
        
        self.assertGreater(a.fuzzy_cmp(Authority("L. Stevens")), 0.5)

    def test_authority_non(self):
        a = Authority("Bert non Miq.")
        self.assertEqual(a.main[0], Author("Bert"))
        self.assertEqual(a.non, (Author("Miq."),))
        self.assertEqual(str(a), "Bert non Miq.")
        
        b = Authority("Miq.")
        assert a.non_match(b)
        assert b.non_match(a)
        self.assertLess(a.fuzzy_cmp(b), 0.2)
    
    def test_authority_extras(self):
        a = Authority("sensu A.Gray")
        self.assertEqual(a.prefix, "sensu")
        self.assertEqual(a.main[0], Author("A. Gray"))
        
        a = Authority("sensu auct., non Benth")
        self.assertEqual(a.prefix, "sensu auct.")
        self.assertEqual(a.non[0], Author("Benth"))
        
        a = Authority("sensu Meeuwen, p.p., non B.L.Burtt")
        self.assertEqual(a.prefix, "sensu")
        self.assertEqual(a.suffix, "p.p.")
        self.assertEqual(a.main[0], Author("Meeuwen"))
        self.assertEqual(a.non[0], Author("B.L. Burtt"))
        
        a = Authority("sensu auct., p.p.")
        self.assertEqual(a.prefix, "sensu auct.")
        self.assertEqual(a.suffix, "p.p.")
    
    def test_with_year(self):
        # Zoological style
        a = Authority("Linnaeus, 1758")
        self.assertEqual(a.main[0], Author("Linnaeus"))
        self.assertEqual(a.year, "1758")
        
        # Without the comma
        a = Authority("Linnaeus 1758")
        self.assertEqual(a.main[0], Author("Linnaeus"))
        self.assertEqual(a.year, "1758")

class AuthorTests(unittest.TestCase):
    def test_surname_f(self):
        a = Author("Forst. f.")
        self.assertEqual(a.to_dict()["surname"], "Forst.")
    
    def test_specialcase_Linnaeus(self):
        self.assertEqual(Author("L."), Author("Linnaeus"))



from .collection import TaxonSet

def combine_datasets(target_ds, background_ds, distrib_ds_name):
    """Combine several taxa datasets into one.
    
    Parameters:
        target_ds, background_ds :
          Each should be a sequence of 2-tuples, (name, taxonset). The name will be
          used as the key in the taxa's info. All taxa from any target dataset are
          included in the output, whereas information from background datasets is
          only copied to taxa in target datasets. There must be at least one target
          dataset, and at least one other (either target or background).
          
          Target datasets may be any iterable collection of taxa. Background datasets
          should be TaxonSets or similar.
        
        distrib_ds_name : str
          The name of the dataset from which distribution information should be
          taken, or 'all' to combine distribution info from all datasets.
    
    Returns: A :class:`TaxonSet`.
    """
    def _check_distrib_ds(dsname):
        return (dsname == distrib_ds_name) or (distrib_ds_name == 'all')

    output = TaxonSet()

    for dsname, ds in target_ds:
        is_distrib_ds = _check_distrib_ds(dsname)
        for tax in ds:
            if tax in output:
                t2 = output[tax.name]
                t2.info[dsname] = tax.info
                if is_distrib_ds:
                    t2.distribution.update(tax.distribution)
            else:
                tax.info = {dsname: tax.info}
                if not is_distrib_ds:
                    tax.distribution = set()
                output.add(tax)

    for dsname, ds in background_ds:
        is_distrib_ds = _check_distrib_ds(dsname)
        for tax in output:
            t2 = ds.get_by_accepted_name(tax.name)
            tax.info[dsname] = t2.info
            if is_distrib_ds:
                tax.distribution.update(t2.distribution)
    
    return output
